/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2012, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TComPic.cpp
    \brief    picture class
*/

#include "TComPic.h"
#include "SEI.h"

//! \ingroup TLibCommon
//! \{

// ====================================================================================================================
// Constructor / destructor / create / destroy
// ====================================================================================================================

TComPic::TComPic()
: m_uiTLayer                              (0)
, m_bUsedByCurr                           (false)
, m_bIsLongTerm                           (false)
, m_bIsUsedAsLongTerm                     (false)
, m_apcPicSym                             (NULL)
, m_pcPicYuvPred                          (NULL)
, m_pcPicYuvResi                          (NULL)
, m_bReconstructed                        (false)
, m_bNeededForOutput                      (false)
, m_uiCurrSliceIdx                        (0)
, m_pSliceSUMap                           (NULL)
, m_pbValidSlice                          (NULL)
, m_sliceGranularityForNDBFilter          (0)
, m_bIndependentSliceBoundaryForNDBFilter (false)
, m_bIndependentTileBoundaryForNDBFilter  (false)
, m_pNDBFilterYuvTmp                      (NULL)
, m_bCheckLTMSB                           (false)
, m_SEIs                                  (NULL)
#if SVC_EXTENSION
, m_bSpatialEnhLayer( false )
, m_pcFullPelBaseRec( NULL )
#if REF_IDX_ME_AROUND_ZEROMV || REF_IDX_MFM
, m_bIsILR                                (false)
#endif
#if REF_IDX_MFM
, m_bIsUpsampledMvField                   (false) 
, m_apcTComUpsampledMvFieldCU             (NULL)
, m_iNumCUInUpsampledPic                  (-1)
#endif
#endif
{
  m_apcPicYuv[0]      = NULL;
  m_apcPicYuv[1]      = NULL;
}

TComPic::~TComPic()
{
}

#if SVC_UPSAMPLING
Void TComPic::create( Int iWidth, Int iHeight, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth, TComSPS* pcSps,  Bool bIsVirtual )
{
  m_apcPicSym     = new TComPicSym;  m_apcPicSym   ->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  if (!bIsVirtual)
  {
    m_apcPicYuv[0]  = new TComPicYuv;  m_apcPicYuv[0]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth, pcSps );
  }
  m_apcPicYuv[1]  = new TComPicYuv;  m_apcPicYuv[1]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth, pcSps );
  
  if (m_bSpatialEnhLayer)
  {
    m_pcFullPelBaseRec = new TComPicYuv;  m_pcFullPelBaseRec->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth, pcSps );
  }

  /* there are no SEI messages associated with this picture initially */
  m_SEIs = NULL;
  m_bUsedByCurr = false;
  return;
}
#if REF_IDX_FRAMEWORK
Void TComPic::createWithOutYuv( Int iWidth, Int iHeight, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth, TComSPS* pcSps,  Bool bIsVirtual)
{
  m_apcPicSym     = new TComPicSym;  m_apcPicSym   ->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  if (!bIsVirtual)
  {
    m_apcPicYuv[0]  = new TComPicYuv;  m_apcPicYuv[0]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth, pcSps );
  }
  m_apcPicYuv[1]  = NULL;
 
#if SVC_UPSAMPLING
  if (m_bSpatialEnhLayer)
  {
    m_pcFullPelBaseRec = new TComPicYuv;  m_pcFullPelBaseRec->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth, pcSps );
  }
#endif

  /* there are no SEI messages associated with this picture initially */
  m_SEIs = NULL;
  m_bUsedByCurr = false;
  return;
}
#endif
#else

Void TComPic::create( Int iWidth, Int iHeight, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth, Bool bIsVirtual )
{
  m_apcPicSym     = new TComPicSym;  m_apcPicSym   ->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  if (!bIsVirtual)
  {
    m_apcPicYuv[0]  = new TComPicYuv;  m_apcPicYuv[0]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  }
  m_apcPicYuv[1]  = new TComPicYuv;  m_apcPicYuv[1]->create( iWidth, iHeight, uiMaxWidth, uiMaxHeight, uiMaxDepth );
  
  /* there are no SEI messages associated with this picture initially */
  m_SEIs = NULL;
  m_bUsedByCurr = false;
  return;
}
#endif

Void TComPic::destroy()
{
  if (m_apcPicSym)
  {
    m_apcPicSym->destroy();
    delete m_apcPicSym;
    m_apcPicSym = NULL;
  }
  
  if (m_apcPicYuv[0])
  {
    m_apcPicYuv[0]->destroy();
    delete m_apcPicYuv[0];
    m_apcPicYuv[0]  = NULL;
  }
  
  if (m_apcPicYuv[1])
  {
    m_apcPicYuv[1]->destroy();
    delete m_apcPicYuv[1];
    m_apcPicYuv[1]  = NULL;
  }
  
#if SVC_EXTENSION && SVC_UPSAMPLING
  if (m_bSpatialEnhLayer)
  {
    m_pcFullPelBaseRec->destroy();
    delete m_pcFullPelBaseRec;    m_pcFullPelBaseRec  = NULL;
  }
#endif 

#if REF_IDX_MFM
  if(m_apcTComUpsampledMvFieldCU)
  { 
    for(Int i = 0; i < m_iNumCUInUpsampledPic; i++) 
	  { 
		  m_apcTComUpsampledMvFieldCU[i]->destroy(); 
		  delete m_apcTComUpsampledMvFieldCU[i]; 
		  m_apcTComUpsampledMvFieldCU[i] = NULL; 
	  } 
	  delete[] m_apcTComUpsampledMvFieldCU; 
	  m_apcTComUpsampledMvFieldCU = NULL; 
  }
#endif

  delete m_SEIs;
}

Void TComPic::compressMotion()
{
  TComPicSym* pPicSym = getPicSym(); 
  for ( UInt uiCUAddr = 0; uiCUAddr < pPicSym->getFrameHeightInCU()*pPicSym->getFrameWidthInCU(); uiCUAddr++ )
  {
    TComDataCU* pcCU = pPicSym->getCU(uiCUAddr);
    pcCU->compressMV(); 
  } 
}

/** Create non-deblocked filter information
 * \param pSliceStartAddress array for storing slice start addresses
 * \param numSlices number of slices in picture
 * \param sliceGranularityDepth slice granularity 
 * \param bNDBFilterCrossSliceBoundary cross-slice-boundary in-loop filtering; true for "cross".
 * \param numTiles number of tiles in picture
 * \param bNDBFilterCrossTileBoundary cross-tile-boundary in-loop filtering; true for "cross".
 */
Void TComPic::createNonDBFilterInfo(std::vector<Int> sliceStartAddress, Int sliceGranularityDepth
                                    ,std::vector<Bool>* LFCrossSliceBoundary
                                    ,Int numTiles
                                    ,Bool bNDBFilterCrossTileBoundary)
{
  UInt maxNumSUInLCU = getNumPartInCU();
  UInt numLCUInPic   = getNumCUsInFrame();
  UInt picWidth      = getSlice(0)->getSPS()->getPicWidthInLumaSamples();
  UInt picHeight     = getSlice(0)->getSPS()->getPicHeightInLumaSamples();
  Int  numLCUsInPicWidth = getFrameWidthInCU();
  Int  numLCUsInPicHeight= getFrameHeightInCU();
  UInt maxNumSUInLCUWidth = getNumPartInWidth();
  UInt maxNumSUInLCUHeight= getNumPartInHeight();
  Int  numSlices = (Int) sliceStartAddress.size() - 1;
  m_bIndependentSliceBoundaryForNDBFilter = false;
  if(numSlices > 1)
  {
    for(Int s=0; s< numSlices; s++)
    {
      if((*LFCrossSliceBoundary)[s] == false)
      {
        m_bIndependentSliceBoundaryForNDBFilter = true;
      }
    }
  }
  m_sliceGranularityForNDBFilter = sliceGranularityDepth;
  m_bIndependentTileBoundaryForNDBFilter  = (bNDBFilterCrossTileBoundary)?(false) :((numTiles > 1)?(true):(false));

  m_pbValidSlice = new Bool[numSlices];
  for(Int s=0; s< numSlices; s++)
  {
    m_pbValidSlice[s] = true;
  }
  m_pSliceSUMap = new Int[maxNumSUInLCU * numLCUInPic];

  //initialization
  for(UInt i=0; i< (maxNumSUInLCU * numLCUInPic); i++ )
  {
    m_pSliceSUMap[i] = -1;
  }
  for( UInt CUAddr = 0; CUAddr < numLCUInPic ; CUAddr++ )
  {
    TComDataCU* pcCU = getCU( CUAddr );
    pcCU->setSliceSUMap(m_pSliceSUMap + (CUAddr* maxNumSUInLCU)); 
    pcCU->getNDBFilterBlocks()->clear();
  }
  m_vSliceCUDataLink.clear();

  m_vSliceCUDataLink.resize(numSlices);

  UInt startAddr, endAddr, firstCUInStartLCU, startLCU, endLCU, lastCUInEndLCU, uiAddr;
  UInt LPelX, TPelY, LCUX, LCUY;
  UInt currSU;
  UInt startSU, endSU;

  for(Int s=0; s< numSlices; s++)
  {
    //1st step: decide the real start address
    startAddr = sliceStartAddress[s];
    endAddr   = sliceStartAddress[s+1] -1;

    startLCU            = startAddr / maxNumSUInLCU;
    firstCUInStartLCU   = startAddr % maxNumSUInLCU;

    endLCU              = endAddr   / maxNumSUInLCU;
    lastCUInEndLCU      = endAddr   % maxNumSUInLCU;   

    uiAddr = m_apcPicSym->getCUOrderMap(startLCU);

    LCUX      = getCU(uiAddr)->getCUPelX();
    LCUY      = getCU(uiAddr)->getCUPelY();
    LPelX     = LCUX + g_auiRasterToPelX[ g_auiZscanToRaster[firstCUInStartLCU] ];
    TPelY     = LCUY + g_auiRasterToPelY[ g_auiZscanToRaster[firstCUInStartLCU] ];
    currSU    = firstCUInStartLCU;

    Bool bMoveToNextLCU = false;
    Bool bSliceInOneLCU = (startLCU == endLCU);

    while(!( LPelX < picWidth ) || !( TPelY < picHeight ))
    {
      currSU ++;

      if(bSliceInOneLCU)
      {
        if(currSU > lastCUInEndLCU)
        {
          m_pbValidSlice[s] = false;
          break;
        }
      }

      if(currSU >= maxNumSUInLCU )
      {
        bMoveToNextLCU = true;
        break;
      }

      LPelX = LCUX + g_auiRasterToPelX[ g_auiZscanToRaster[currSU] ];
      TPelY = LCUY + g_auiRasterToPelY[ g_auiZscanToRaster[currSU] ];

    }


    if(!m_pbValidSlice[s])
    {
      continue;
    }

    if(currSU != firstCUInStartLCU)
    {
      if(!bMoveToNextLCU)
      {
        firstCUInStartLCU = currSU;
      }
      else
      {
        startLCU++;
        firstCUInStartLCU = 0;
        assert( startLCU < getNumCUsInFrame());
      }
      assert(startLCU*maxNumSUInLCU + firstCUInStartLCU < endAddr);
    }


    //2nd step: assign NonDBFilterInfo to each processing block
    for(UInt i= startLCU; i <= endLCU; i++)
    {
      startSU = (i == startLCU)?(firstCUInStartLCU):(0);
      endSU   = (i == endLCU  )?(lastCUInEndLCU   ):(maxNumSUInLCU -1);

      uiAddr = m_apcPicSym->getCUOrderMap(i);
      Int iTileID= m_apcPicSym->getTileIdxMap(uiAddr);

      TComDataCU* pcCU = getCU(uiAddr);
      m_vSliceCUDataLink[s].push_back(pcCU);

      createNonDBFilterInfoLCU(iTileID, s, pcCU, startSU, endSU, m_sliceGranularityForNDBFilter, picWidth, picHeight);
    }
  }

  //step 3: border availability
  for(Int s=0; s< numSlices; s++)
  {
    if(!m_pbValidSlice[s])
    {
      continue;
    }

    for(Int i=0; i< m_vSliceCUDataLink[s].size(); i++)
    {
      TComDataCU* pcCU = m_vSliceCUDataLink[s][i];
      uiAddr = pcCU->getAddr();

      if(pcCU->getPic()==0)
      {
        continue;
      }
      Int iTileID= m_apcPicSym->getTileIdxMap(uiAddr);
      Bool bTopTileBoundary = false, bDownTileBoundary= false, bLeftTileBoundary= false, bRightTileBoundary= false;

      if(m_bIndependentTileBoundaryForNDBFilter)
      {
        //left
        if( uiAddr % numLCUsInPicWidth != 0)
        {
          bLeftTileBoundary = ( m_apcPicSym->getTileIdxMap(uiAddr -1) != iTileID )?true:false;
        }
        //right
        if( (uiAddr % numLCUsInPicWidth) != (numLCUsInPicWidth -1) )
        {
          bRightTileBoundary = ( m_apcPicSym->getTileIdxMap(uiAddr +1) != iTileID)?true:false;
        }
        //top
        if( uiAddr >= numLCUsInPicWidth)
        {
          bTopTileBoundary = (m_apcPicSym->getTileIdxMap(uiAddr - numLCUsInPicWidth) !=  iTileID )?true:false;
        }
        //down
        if( uiAddr + numLCUsInPicWidth < numLCUInPic )
        {
          bDownTileBoundary = (m_apcPicSym->getTileIdxMap(uiAddr + numLCUsInPicWidth) != iTileID)?true:false;
        }

      }

      pcCU->setNDBFilterBlockBorderAvailability(numLCUsInPicWidth, numLCUsInPicHeight, maxNumSUInLCUWidth, maxNumSUInLCUHeight,picWidth, picHeight
        , *LFCrossSliceBoundary
        ,bTopTileBoundary, bDownTileBoundary, bLeftTileBoundary, bRightTileBoundary
        ,m_bIndependentTileBoundaryForNDBFilter);

    }

  }

  if( m_bIndependentSliceBoundaryForNDBFilter || m_bIndependentTileBoundaryForNDBFilter)
  {
    m_pNDBFilterYuvTmp = new TComPicYuv();
    m_pNDBFilterYuvTmp->create(picWidth, picHeight, g_uiMaxCUWidth, g_uiMaxCUHeight, g_uiMaxCUDepth);
  }

}

/** Create non-deblocked filter information for LCU
 * \param tileID tile index
 * \param sliceID slice index
 * \param pcCU CU data pointer
 * \param startSU start SU index in LCU
 * \param endSU end SU index in LCU
 * \param sliceGranularyDepth slice granularity
 * \param picWidth picture width
 * \param picHeight picture height
 */
Void TComPic::createNonDBFilterInfoLCU(Int tileID, Int sliceID, TComDataCU* pcCU, UInt startSU, UInt endSU, Int sliceGranularyDepth, UInt picWidth, UInt picHeight)
{
  UInt LCUX          = pcCU->getCUPelX();
  UInt LCUY          = pcCU->getCUPelY();
  Int* pCUSliceMap    = pcCU->getSliceSUMap();
  UInt maxNumSUInLCU = getNumPartInCU();
  UInt maxNumSUInSGU = maxNumSUInLCU >> (sliceGranularyDepth << 1);
  UInt maxNumSUInLCUWidth = getNumPartInWidth();
  UInt LPelX, TPelY;
  UInt currSU;


  //get the number of valid NBFilterBLock
  currSU   = startSU;
  while(currSU <= endSU)
  {
    LPelX = LCUX + g_auiRasterToPelX[ g_auiZscanToRaster[currSU] ];
    TPelY = LCUY + g_auiRasterToPelY[ g_auiZscanToRaster[currSU] ];

    while(!( LPelX < picWidth ) || !( TPelY < picHeight ))
    {
      currSU += maxNumSUInSGU;
      if(currSU >= maxNumSUInLCU || currSU > endSU)
      {
        break;
      }
      LPelX = LCUX + g_auiRasterToPelX[ g_auiZscanToRaster[currSU] ];
      TPelY = LCUY + g_auiRasterToPelY[ g_auiZscanToRaster[currSU] ];
    }

    if(currSU >= maxNumSUInLCU || currSU > endSU)
    {
      break;
    }

    NDBFBlockInfo NDBFBlock;

    NDBFBlock.tileID  = tileID;
    NDBFBlock.sliceID = sliceID;
    NDBFBlock.posY    = TPelY;
    NDBFBlock.posX    = LPelX;
    NDBFBlock.startSU = currSU;

    UInt uiLastValidSU  = currSU;
    UInt uiIdx, uiLPelX_su, uiTPelY_su;
    for(uiIdx = currSU; uiIdx < currSU + maxNumSUInSGU; uiIdx++)
    {
      if(uiIdx > endSU)
      {
        break;        
      }
      uiLPelX_su   = LCUX + g_auiRasterToPelX[ g_auiZscanToRaster[uiIdx] ];
      uiTPelY_su   = LCUY + g_auiRasterToPelY[ g_auiZscanToRaster[uiIdx] ];
      if( !(uiLPelX_su < picWidth ) || !( uiTPelY_su < picHeight ))
      {
        continue;
      }
      pCUSliceMap[uiIdx] = sliceID;
      uiLastValidSU = uiIdx;
    }
    NDBFBlock.endSU = uiLastValidSU;

    UInt rTLSU = g_auiZscanToRaster[ NDBFBlock.startSU ];
    UInt rBRSU = g_auiZscanToRaster[ NDBFBlock.endSU   ];
    NDBFBlock.widthSU  = (rBRSU % maxNumSUInLCUWidth) - (rTLSU % maxNumSUInLCUWidth)+ 1;
    NDBFBlock.heightSU = (UInt)(rBRSU / maxNumSUInLCUWidth) - (UInt)(rTLSU / maxNumSUInLCUWidth)+ 1;
    NDBFBlock.width    = NDBFBlock.widthSU  * getMinCUWidth();
    NDBFBlock.height   = NDBFBlock.heightSU * getMinCUHeight();

    pcCU->getNDBFilterBlocks()->push_back(NDBFBlock);

    currSU += maxNumSUInSGU;
  }

}

/** destroy non-deblocked filter information for LCU
 */
Void TComPic::destroyNonDBFilterInfo()
{
  if(m_pbValidSlice != NULL)
  {
    delete[] m_pbValidSlice;
    m_pbValidSlice = NULL;
  }

  if(m_pSliceSUMap != NULL)
  {
    delete[] m_pSliceSUMap;
    m_pSliceSUMap = NULL;
  }
  for( UInt CUAddr = 0; CUAddr < getNumCUsInFrame() ; CUAddr++ )
  {
    TComDataCU* pcCU = getCU( CUAddr );
    pcCU->getNDBFilterBlocks()->clear();
  }

  if( m_bIndependentSliceBoundaryForNDBFilter || m_bIndependentTileBoundaryForNDBFilter)
  {
    m_pNDBFilterYuvTmp->destroy();
    delete m_pNDBFilterYuvTmp;
    m_pNDBFilterYuvTmp = NULL;
  }

}

#if REF_IDX_FRAMEWORK
Void copyOnetoOnePicture(    // SVC_NONCOLL
                  Pel *in,        
                  Pel *out,      
                  Int nCols,
                  Int nRows, 
                  Int fullRowWidth)
{
  Int rX;

  for (rX = 0; rX < nRows; rX++)       
  {
    memcpy( out, in, sizeof(Pel) * nCols );
    in = in + fullRowWidth;
    out = out + fullRowWidth;
  }
}

Void TComPic:: copyUpsampledPictureYuv(TComPicYuv*   pcPicYuvIn, TComPicYuv*   pcPicYuvOut)
{
  Int upsampledRowWidthLuma = pcPicYuvOut->getStride(); // 2 * pcPicYuvOut->getLumaMargin() + pcPicYuvOut->getWidth(); 
  Int upsampledRowWidthCroma = pcPicYuvOut->getCStride(); //2 * pcPicYuvOut->getChromaMargin() + (pcPicYuvOut->getWidth()>>1);

  copyOnetoOnePicture(
    pcPicYuvIn->getLumaAddr(),        
    pcPicYuvOut->getLumaAddr(),      
    pcPicYuvOut->getWidth(), 
    pcPicYuvOut->getHeight(),
    upsampledRowWidthLuma);
  copyOnetoOnePicture(
    pcPicYuvIn->getCrAddr(),        
    pcPicYuvOut->getCrAddr(),      
    pcPicYuvOut->getWidth()>>1, 
    pcPicYuvOut->getHeight()>>1,
    upsampledRowWidthCroma);
  copyOnetoOnePicture(
    pcPicYuvIn->getCbAddr(),        
    pcPicYuvOut->getCbAddr(),      
    pcPicYuvOut->getWidth()>>1, 
    pcPicYuvOut->getHeight()>>1,
    upsampledRowWidthCroma);
}
#endif

#if REF_IDX_MFM
Void TComPic::createUpSampledMvField(Int upSampledHeight, Int upSampledWidth, UInt uiMaxWidth, UInt uiMaxHeight, UInt uiMaxDepth )
{
	UInt uiNumPartitions   = 1<<(uiMaxDepth<<1);
	UInt uiWidthInCU       = ( upSampledWidth % uiMaxWidth  ) ? upSampledWidth /uiMaxWidth  + 1 : upSampledWidth /uiMaxWidth;
  UInt uiHeightInCU      = ( upSampledHeight% uiMaxHeight ) ? upSampledHeight/uiMaxHeight + 1 : upSampledHeight/uiMaxHeight;
	UInt uiNumCUsInFrame   = uiWidthInCU * uiHeightInCU;
	m_iNumCUInUpsampledPic = uiNumCUsInFrame;

	m_apcTComUpsampledMvFieldCU     = new TComUpsampledMvFieldCU*[uiNumCUsInFrame];

	for (Int i = 0; i < uiNumCUsInFrame; i++ )
	{
		m_apcTComUpsampledMvFieldCU[i] = new TComUpsampledMvFieldCU;
		m_apcTComUpsampledMvFieldCU[i]->createMvField(i, uiWidthInCU, uiNumPartitions);  //create Mv field
		m_apcTComUpsampledMvFieldCU[i]->initMvField();                                   //initialize Mv field
	}

	return;
}

Void TComPic::deriveUnitIdxBase(UInt uiUpsamplePelX, UInt uiUpsamplePelY, float ratio, UInt& uiBaseCUAddr, UInt& uiBaseAbsPartIdx)
{
  //pixel in the base layer
  UInt uiPelX       = (UInt)((float)uiUpsamplePelX/ratio);
  UInt uiPelY       = (UInt)((float)uiUpsamplePelY/ratio);
  UInt uiBaseWidth  = getPicYuvRec()->getWidth();
  UInt uiBaseHeight = getPicYuvRec()->getHeight();

  UInt uiWidthInCU       = ( uiBaseWidth % g_uiMaxCUWidth  ) ? uiBaseWidth /g_uiMaxCUWidth  + 1 : uiBaseWidth /g_uiMaxCUWidth;
  UInt uiHeightInCU      = ( uiBaseHeight% g_uiMaxCUHeight ) ? uiBaseHeight/ g_uiMaxCUHeight + 1 : uiBaseHeight/ g_uiMaxCUHeight;

  uiPelX     = (UInt)Clip3<UInt>(0, uiWidthInCU * g_uiMaxCUWidth - 1, uiPelX);
  uiPelY     = (UInt)Clip3<UInt>(0, uiHeightInCU * g_uiMaxCUHeight - 1, uiPelY);
  
  uiBaseCUAddr = uiPelY / g_uiMaxCUHeight * uiWidthInCU + uiPelX / g_uiMaxCUWidth;

  UInt uiWidthMinPU = g_uiMaxCUWidth / (1<<g_uiMaxCUDepth);
  UInt uiHeightMinPU = g_uiMaxCUHeight/(1<<g_uiMaxCUDepth);
  
  UInt uiAbsPelX = uiPelX - (uiPelX / g_uiMaxCUWidth) * g_uiMaxCUWidth;
  UInt uiAbsPelY = uiPelY - (uiPelY / g_uiMaxCUHeight) * g_uiMaxCUHeight;

  UInt RasterIdx = uiAbsPelY / uiHeightMinPU * (g_uiMaxCUWidth/uiWidthMinPU) + uiAbsPelX / uiWidthMinPU;
  uiBaseAbsPartIdx = g_auiRasterToZscan[RasterIdx];

  return;
}

//check if identical unit already exist
Bool TComPic::xCheckIdenticalCUAddr(UInt uiBaseCUAddr, UInt uiBaseAbsPartIdx, UInt* BaseCUAddrCands, UInt* AbsPartIdxCands, UInt numValidCands)
{
  assert(numValidCands <= 4);
  if(numValidCands == 0) return false;

  Bool found = false;
  for(Int i = 0; i < numValidCands; i++)
  {
    if(BaseCUAddrCands[i] == uiBaseCUAddr && AbsPartIdxCands[i] == uiBaseAbsPartIdx)
    {
      found = true;
      break;
    }
  }

  return found;
}

//select reference index of units in up-sampled layer
template <typename T> inline T Max( T a, T b)            { return std::max<T> (a, b); }
template <typename T> inline T Sum3( T a, T b, T c)      { return a + b + c; }
template <typename T> inline T Max3( T a, T b, T c)      { return std::max<T> (std::max<T> (a, b) , c); }
template <typename T> inline T Min3( T a, T b, T c)      { return std::min<T> (std::min<T> (a, b) , c); }
template <typename T> inline T Med3( T a, T b, T c)      { return Sum3<T>(a, b, c) - Max3<T>(a, b, c) - Min3<T>(a, b, c); }
template <typename T> inline T Max4( T a, T b, T c, T d) { return std::max<T> (std::max<T> (a, b) , std::max<T> (c, d)); }

Int TComPic::decideRefIdx(UInt* BaseCUAddrCands, UInt* AbsPartIdxCands, UInt numValidCands, RefPicList refPicList)
{
  Int  refIdxCands[4];
  UInt freqUsage[4];

  assert(numValidCands <= 4);

  UInt numInterCands = 0;
  for(Int cIdx = 0; cIdx < numValidCands; cIdx++)
    if(getCU(BaseCUAddrCands[cIdx])->getCUMvField(refPicList)->getRefIdx(AbsPartIdxCands[cIdx]) >= 0)  //not intra coded
      refIdxCands[numInterCands++] = getCU(BaseCUAddrCands[cIdx])->getCUMvField(refPicList)->getRefIdx(AbsPartIdxCands[cIdx]);

  //check the frequency usage
  memset(freqUsage, 0, 4 * sizeof(UInt));
  for(Int cIdx = 0; cIdx < numInterCands; cIdx++)
  {
    for(Int sIdx = 0; sIdx < numInterCands; sIdx++) 
      if(refIdxCands[sIdx] == refIdxCands[cIdx]) 
        freqUsage[cIdx]++;
  }

  UInt freq_max;
  if(numInterCands == 0)  //all intra coded
    return NOT_VALID;
  else if(numInterCands == 1)  //one inter candidate
    return refIdxCands[0];
  else if(numInterCands == 2)  //two inter candidates 
    freq_max = (UInt)Max<UInt>(freqUsage[0], freqUsage[1]);
  else if(numInterCands == 3)  //three inter candidates
    freq_max = (UInt)Max3<UInt>(freqUsage[0], freqUsage[1], freqUsage[2]);
  else if(numInterCands == 4)  //four inter candidates
    freq_max = (UInt)Max4<UInt>(freqUsage[0], freqUsage[1], freqUsage[2], freqUsage[3]);

  Int selectedRefIdx = MAX_INT;
  for(Int cIdx = 0; cIdx < numInterCands; cIdx++)
  {
    if(freqUsage[cIdx] == freq_max && refIdxCands[cIdx] < selectedRefIdx)
      selectedRefIdx = refIdxCands[cIdx];
  }
  
  return selectedRefIdx;
}

//MV field up-sampling
Void TComPic::doTheUpSampleMvField(UInt upSampleRatio)
{
  UInt uiWidthMinPU      = g_uiMaxCUWidth/(1<<g_uiMaxCUDepth);
  UInt uiHeightMinPU     = g_uiMaxCUHeight/(1<<g_uiMaxCUDepth);
  UInt uiNumPartitions   = 1<<(g_uiMaxCUDepth<<1);
  UInt uiPelX, uiPelY;
  UInt uiBaseCUAddr, uiBaseAbsPartIdx;
  float ratio;

  if(upSampleRatio == 0)       //ONE_TO_ONE
    ratio = 1.0;
  else if(upSampleRatio == 1)  //ONE_TO_TWO
    ratio = 2.0;
  else if(upSampleRatio == 2)  //TWO_TO_THREE
    ratio = 1.5;

  for(UInt uiCUAddr = 0; uiCUAddr < m_iNumCUInUpsampledPic; uiCUAddr++)  //each LCU
  {
    TComUpsampledMvFieldCU*& pcMvFieldCU = getUpsampledMvFieldCU(uiCUAddr);
    pcMvFieldCU->initMvField();                                          //MV field initialization
    
    for(UInt uiAbsPartIdx = 0; uiAbsPartIdx < uiNumPartitions; uiAbsPartIdx++)  //each smallest unit
    {
      UInt        BaseCUAddrCands[4];                                  //CU address of composing units in base layer
      UInt        AbsPartIdxCands[4];                                  //unit index of composing units in base layer
      UInt        numValidCands = 0;                                   //number of valid composing units in base layer

      //pixel position of each unit in up-sampled layer
      uiPelX = pcMvFieldCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsPartIdx] ];
      uiPelY = pcMvFieldCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsPartIdx] ];

      //left-top units in the base layer
      deriveUnitIdxBase(uiPelX, uiPelY, ratio, uiBaseCUAddr, uiBaseAbsPartIdx);      
      if( (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_NONE) && (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_INTRA) )  //base layer unit not intra and invalid mode
      {
        if(!xCheckIdenticalCUAddr(uiBaseCUAddr, uiBaseAbsPartIdx, BaseCUAddrCands, AbsPartIdxCands, numValidCands))
        {
          BaseCUAddrCands[numValidCands] = uiBaseCUAddr;
          AbsPartIdxCands[numValidCands] = uiBaseAbsPartIdx;
          numValidCands++;
        }
      }

      //right-top units in the base layer
      deriveUnitIdxBase(uiPelX + uiWidthMinPU -1, uiPelY, ratio, uiBaseCUAddr, uiBaseAbsPartIdx);
      if( (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_NONE) && (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_INTRA) )  //base layer unit not intra and invalid mode
      {
        if(!xCheckIdenticalCUAddr(uiBaseCUAddr, uiBaseAbsPartIdx, BaseCUAddrCands, AbsPartIdxCands, numValidCands))
        {
          BaseCUAddrCands[numValidCands] = uiBaseCUAddr;
          AbsPartIdxCands[numValidCands] = uiBaseAbsPartIdx;
          numValidCands++;
        }
      }

      //left-bottom units in the base layer
      deriveUnitIdxBase(uiPelX, uiPelY + uiHeightMinPU - 1, ratio, uiBaseCUAddr, uiBaseAbsPartIdx);
      if( (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_NONE) && (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_INTRA) )  //base layer unit not intra and invalid mode
      {
        if(!xCheckIdenticalCUAddr(uiBaseCUAddr, uiBaseAbsPartIdx, BaseCUAddrCands, AbsPartIdxCands, numValidCands))
        {
          BaseCUAddrCands[numValidCands] = uiBaseCUAddr;
          AbsPartIdxCands[numValidCands] = uiBaseAbsPartIdx;
          numValidCands++;
        }
      }

      //right-bottom units in the base layer
      deriveUnitIdxBase(uiPelX + uiWidthMinPU -1, uiPelY + uiHeightMinPU - 1, ratio, uiBaseCUAddr, uiBaseAbsPartIdx);
      if( (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_NONE) && (getCU(uiBaseCUAddr)->getPredictionMode(uiBaseAbsPartIdx) != MODE_INTRA) )  //base layer unit not skip and invalid mode
      {
        if(!xCheckIdenticalCUAddr(uiBaseCUAddr, uiBaseAbsPartIdx, BaseCUAddrCands, AbsPartIdxCands, numValidCands))
        {
          BaseCUAddrCands[numValidCands] = uiBaseCUAddr;
          AbsPartIdxCands[numValidCands] = uiBaseAbsPartIdx;
          numValidCands++;
        }
      }

      if(numValidCands == 0)  //if there is no valid base layer composing unit
      {
        TComMvField zeroMvField;  //zero MV and invalid reference index
        pcMvFieldCU->getCUMvField(REF_PIC_LIST_0)->setMvField(zeroMvField, uiAbsPartIdx);
        pcMvFieldCU->getCUMvField(REF_PIC_LIST_1)->setMvField(zeroMvField, uiAbsPartIdx);
      }
      else  //if there is at least one valid base layer composing unit
      {
        for(UInt list = 0; list < 2; list++)  //each list
        { 
          Int selectedRefIdx = decideRefIdx(BaseCUAddrCands, AbsPartIdxCands, numValidCands, (RefPicList)list);
          if(selectedRefIdx == NOT_VALID)
          {
            TComMvField zeroMvField;
            pcMvFieldCU->getCUMvField((RefPicList)list)->setMvField(zeroMvField, uiAbsPartIdx);
          }
          else
          {
            TComMv cMvCands[4];          //candidate MVs
            UInt   numMvCands = 0;
            for(Int cIdx = 0; cIdx < numValidCands; cIdx++)
              if(getCU(BaseCUAddrCands[cIdx])->getCUMvField((RefPicList)list)->getRefIdx(AbsPartIdxCands[cIdx]) == selectedRefIdx)
              {
                TComMv cMv = getCU(BaseCUAddrCands[cIdx])->getCUMvField((RefPicList)list)->getMv(AbsPartIdxCands[cIdx]);
                cMvCands[numMvCands++].set(cMv.getHor(), cMv.getVer());
              }

            assert(numMvCands > 0);
            TComMvField sMvField;
            if(numMvCands == 1)  //one MV candidate case
            {
              Short Hor = cMvCands[0].getHor();
              Short Ver = cMvCands[0].getVer();
              Hor = (Short) (ratio * Hor);
              Ver = (Short) (ratio * Ver);

              TComMv cMv(Hor, Ver);
              sMvField.setMvField(cMv, selectedRefIdx);
            }
            else if(numMvCands == 2)  //two MV candidates case
            {
              Short Hor = (cMvCands[0].getHor()+cMvCands[1].getHor())/2;
              Short Ver = (cMvCands[0].getVer()+cMvCands[1].getVer())/2;
              Hor = (Short) (ratio * Hor);
              Ver = (Short) (ratio * Ver);
              
              TComMv cMv(Hor, Ver);
              sMvField.setMvField(cMv, selectedRefIdx);
            }
            else if(numMvCands == 3)  //three MV candidates case
            {
              Short Hor = Med3<Short>(cMvCands[0].getHor(), cMvCands[1].getHor(), cMvCands[2].getHor());
              Short Ver = Med3<Short>(cMvCands[0].getVer(), cMvCands[1].getVer(), cMvCands[2].getVer());
              Hor = (Short) (ratio * Hor);
              Ver = (Short) (ratio * Ver);

              TComMv cMv(Hor, Ver);
              sMvField.setMvField(cMv, selectedRefIdx);
            }
            else if(numMvCands == 4)  //four MV candidates case
            {
              Short Hor = Med3<Short>(cMvCands[0].getHor(), cMvCands[1].getHor(), cMvCands[2].getHor());
              Hor += Med3<Short>(cMvCands[1].getHor(), cMvCands[2].getHor(), cMvCands[3].getHor());
              Hor = Hor/2;
              Short Ver = Med3<Short>(cMvCands[0].getVer(), cMvCands[1].getVer(), cMvCands[2].getVer());
              Ver += Med3<Short>(cMvCands[1].getVer(), cMvCands[2].getVer(), cMvCands[3].getVer());
              Ver = Ver/2;
              Hor = (Short) (ratio * Hor);
              Ver = (Short) (ratio * Ver);

              TComMv cMv(Hor, Ver);
              sMvField.setMvField(cMv, selectedRefIdx);
            }

            pcMvFieldCU->getCUMvField((RefPicList)list)->setMvField(sMvField, uiAbsPartIdx);
          }  //ending else
        }  //ending for(UInt list = 0; list < 2; list++)
      }  //ending else
    }  //ending for(UInt uiAbsPartIdx = 0; uiAbsPartIdx < uiNumPartitions; uiAbsPartIdx++) 
  }  //ending for(UInt uiCUAddr = 0; uiCUAddr < m_iNumCUInUpsampledPic; uiCUAddr++)
  return;
}

Void TComPic::copyUpsampledMvField(TComPic* pcPicBase)
{
  assert( getPicSym()->getNumberOfCUsInFrame() == pcPicBase->getNumCUInUpsampledPic() );

  for(UInt cuIdx = 0; cuIdx < getPicSym()->getNumberOfCUsInFrame(); cuIdx++)  //each LCU
  {
    TComDataCU*             pcCUDes = getCU(cuIdx);
    TComUpsampledMvFieldCU* pcMvFieldCU = pcPicBase->getUpsampledMvFieldCU(cuIdx);
    for(UInt list = 0; list < 2; list++)  //each reference list
    {
      TComCUMvField* pcCUMvFieldDes = pcCUDes->getCUMvField((RefPicList)list);
      TComCUMvField* pcCUMvFieldOrg = pcMvFieldCU->getCUMvField((RefPicList)list);

      assert(pcCUMvFieldDes->getNumPartition() == pcCUMvFieldOrg->getNumPartition());
      pcCUMvFieldDes->clearMvField();
      pcCUMvFieldDes->copyFrom(pcCUMvFieldOrg, pcCUMvFieldDes->getNumPartition(), 0);      
    }
  }
}
#endif

//! \}
